#include "filter_proj_float.h"

// static float gains[8] = {1, 1, 1, 1, 1, 1, 1, 1};
// static float gains[8] = {-10, -10, -10, 0, 0, 10, 10, 10};
// static float center_frequencies[8] = {125, 250, 500, 1000, 2000, 4000, 8000, 14000};
// static float band_widths[8] = {1.0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 1.0};
// static float band_widths[8] = {1.0, 0.75, 0.75, 0.75, 0.75, 0.75, 0.75, 1.0};

void normalize(float* audio_sig)
{
    float maxo = -1;
    int i = 0;
    for (i = 0; i < audio_len; ++i)
    {
        if (fabs(audio_sig[i]) > maxo)
            maxo = fabs(audio_sig[i]);
    }
    for (i = 0; i < audio_len; ++i)
    {
        audio_sig[i] = audio_sig[i]/maxo;
    }
}

void create_filter_fast(float* B, float* A, arm_fir_instance_f32* filt, float pState[fft_size*2-1])
{
	arm_rfft_fast_instance_f32 varInstRfftF32;
	arm_rfft_fast_init_f32(&varInstRfftF32, fft_size);

    float eps = 1e-6;
    int i = 0;

    float complex fd_B_out[n_freqs];
    float complex fd_A_out[n_freqs];
    float complex tempo[n_freqs];
    arm_rfft_fast_f32(&varInstRfftF32, (float*)B, (float*)fd_B_out, 0);
    arm_rfft_fast_f32(&varInstRfftF32, (float*)A, (float*)fd_A_out, 0);

    ((float*)(fd_B_out))[fft_size] =  ((float*)(fd_B_out))[1];
    ((float*)(fd_B_out))[fft_size+1] = 0;
    ((float*)(fd_B_out))[1] = 0;

    ((float*)(fd_A_out))[fft_size] =  ((float*)(fd_A_out))[1];
    ((float*)(fd_A_out))[fft_size+1] = 0;
    ((float*)(fd_A_out))[1] = 0;

    for (i = 0; i < n_freqs; ++i)
    {
        tempo[i] = fd_B_out[i]/(fd_A_out[i]+eps);
    }
    arm_rfft_fast_f32(&varInstRfftF32, (float*)tempo, (float*)B, 1);
    for (i = 0; i < fft_size; ++i)
    	A[fft_size-1-i] = B[i];
    arm_fir_init_f32(filt, fft_size, A, pState, fft_size);
}

void create_filter(float* B, float* A, float complex* filt)
{
	arm_rfft_fast_instance_f32 varInstRfftF32;
	arm_rfft_fast_init_f32(&varInstRfftF32, fft_size);

    float eps = 1e-6;
    int i = 0;

    // compute the FFT of B
    // float fd_B[fft_size];
    // float fd_A[fft_size];
    float complex fd_B_out[n_freqs];
    float complex fd_A_out[n_freqs];
    /*
    for (i = 0; i < fft_size; ++i)
    {
        fd_B[i] = B[i];
        fd_A[i] = A[i];
    }
    */
    // fft(fd_B, fft_size);
    // fft(fd_A, fft_size);
    arm_rfft_fast_f32(&varInstRfftF32, (float*)B, (float*)fd_B_out, 0);
    arm_rfft_fast_f32(&varInstRfftF32, (float*)A, (float*)fd_A_out, 0);

    ((float*)(fd_B_out))[fft_size] =  ((float*)(fd_B_out))[1];
    ((float*)(fd_B_out))[fft_size+1] = 0;
    ((float*)(fd_B_out))[1] = 0;

    ((float*)(fd_A_out))[fft_size] =  ((float*)(fd_A_out))[1];
    ((float*)(fd_A_out))[fft_size+1] = 0;
    ((float*)(fd_A_out))[1] = 0;

    for (i = 0; i < n_freqs; ++i)
    {
        filt[i] = fd_B_out[i]/(fd_A_out[i]+eps);
    	// filt[i] = fd_B[i]/(fd_A[i]+eps);
    }

}

float fn_sinh(float x)
{
    return 0.5*(exp(x)-exp(-1*x));
}

void true_bell_filter_fast(float K, float fc, float BW, float fs, arm_fir_instance_f32* freq_filter, float pState[fft_size*2-1])
{
    float A = pow(10, (K/20));
    float w0 = 2*pi*fc/fs;
    float alpha = sin(w0)*fn_sinh(log(2)/2*BW*w0/sin(w0));

    float b0 = 1 + alpha*A;
    float b1 = -2 * cos(w0);
    float b2 = 1 - alpha*A;
    float a0 = 1 + alpha/A;
    float a1 = -2 * cos(w0);
    float a2 = 1 - alpha/A;

    float A_[fft_size];
    float B_[fft_size];
    int i = 0;
    A_[0] = a0;
    A_[1] = a1;
    A_[2] = a2;
    B_[0] = b0;
    B_[1] = b1;
    B_[2] = b2;
    arm_fill_f32(0, A_+3, (fft_size-3));
    arm_fill_f32(0, B_+3, (fft_size-3));
    /*
    for (i = 3; i < fft_size; ++i)
    {
        A_[i] = 0;
        B_[i] = 0;
    }
    */
    create_filter_fast(B_, A_, freq_filter, pState);
}

void low_shelf_fast(float K, float fc, float S, float fs, arm_fir_instance_f32* freq_filter, float pState[fft_size*2-1])
{
    float A = pow(10, (K/20));
    float w0 = 2*pi*fc/fs;
    float alpha = sin(w0)/2 * sqrt((A+1/A)*(1/S-1)+2);

    float b0 = A*((A+1)-(A-1)*cos(w0)+2*sqrt(A)*alpha);
    float b1 = 2*A*((A-1)-(A+1)*cos(w0));
    float b2 = A*((A+1)-(A-1)*cos(w0)-2*sqrt(A)*alpha);
    float a0 = (A+1)+(A-1)*cos(w0)+2*sqrt(A)*alpha;
    float a1 = -2*((A-1)+(A+1)*cos(w0));
    float a2 = (A+1)+(A-1)*cos(w0)-2*sqrt(A)*alpha;

    float A_[fft_size];
    float B_[fft_size];
    int i = 0;
    A_[0] = a0;
    A_[1] = a1;
    A_[2] = a2;
    B_[0] = b0;
    B_[1] = b1;
    B_[2] = b2;
    arm_fill_f32(0, A_+3, (fft_size-3));
    arm_fill_f32(0, B_+3, (fft_size-3));
    /*
    for (i = 3; i < fft_size; ++i)
    {
        A_[i] = 0;
        B_[i] = 0;
    }
    */
    create_filter_fast(B_, A_, freq_filter, pState);
}

void high_shelf_fast(float K, float fc, float S, float fs, arm_fir_instance_f32* freq_filter, float pState[fft_size*2-1])
{
    float A = pow(10, (K/20));
    float w0 = 2*pi*fc/fs;
    float alpha = sin(w0)/2 * sqrt((A+1/A)*(1/S-1)+2);

    float b0 = A*((A+1)+(A-1)*cos(w0)+2*sqrt(A)*alpha);
    float b1 = -2*A*((A-1)+(A+1)*cos(w0));
    float b2 = A*((A+1)+(A-1)*cos(w0)-2*sqrt(A)*alpha);
    float a0 = (A+1)-(A-1)*cos(w0)+2*sqrt(A)*alpha;
    float a1 = 2*((A-1)-(A+1)*cos(w0));
    float a2 = (A+1)-(A-1)*cos(w0)-2*sqrt(A)*alpha;

    float A_[fft_size];
    float B_[fft_size];
    int i = 0;
    A_[0] = a0;
    A_[1] = a1;
    A_[2] = a2;
    B_[0] = b0;
    B_[1] = b1;
    B_[2] = b2;
    arm_fill_f32(0, A_+3, (fft_size-3));
    arm_fill_f32(0, B_+3, (fft_size-3));
    /*
    for (i = 3; i < fft_size; ++i)
    {
        A_[i] = 0;
        B_[i] = 0;
    }
    */
    create_filter_fast(B_, A_, freq_filter, pState);
}

void true_bell_filter(float K, float fc, float BW, float fs, float complex* freq_filter)
{
    float A = pow(10, (K/20));
    float w0 = 2*pi*fc/fs;
    float alpha = sin(w0)*fn_sinh(log(2)/2*BW*w0/sin(w0));
    
    float b0 = 1 + alpha*A;
    float b1 = -2 * cos(w0);
    float b2 = 1 - alpha*A;
    float a0 = 1 + alpha/A;
    float a1 = -2 * cos(w0);
    float a2 = 1 - alpha/A;
    
    float A_[fft_size];
    float B_[fft_size];
    int i = 0;
    A_[0] = a0;
    A_[1] = a1;
    A_[2] = a2;
    B_[0] = b0;
    B_[1] = b1;
    B_[2] = b2;
    arm_fill_f32(0, A_+3, (fft_size-3));
    arm_fill_f32(0, B_+3, (fft_size-3));
    /*
    for (i = 3; i < fft_size; ++i)
    {
        A_[i] = 0;
        B_[i] = 0;
    }
    */
    create_filter(B_, A_, freq_filter);
}

void low_shelf(float K, float fc, float S, float fs, float complex* freq_filter)
{
    float A = pow(10, (K/20));
    float w0 = 2*pi*fc/fs;
    float alpha = sin(w0)/2 * sqrt((A+1/A)*(1/S-1)+2);
    
    float b0 = A*((A+1)-(A-1)*cos(w0)+2*sqrt(A)*alpha);
    float b1 = 2*A*((A-1)-(A+1)*cos(w0));
    float b2 = A*((A+1)-(A-1)*cos(w0)-2*sqrt(A)*alpha);
    float a0 = (A+1)+(A-1)*cos(w0)+2*sqrt(A)*alpha;
    float a1 = -2*((A-1)+(A+1)*cos(w0));
    float a2 = (A+1)+(A-1)*cos(w0)-2*sqrt(A)*alpha;
    
    float A_[fft_size];
    float B_[fft_size];
    int i = 0;
    A_[0] = a0;
    A_[1] = a1;
    A_[2] = a2;
    B_[0] = b0;
    B_[1] = b1;
    B_[2] = b2;
    arm_fill_f32(0, A_+3, (fft_size-3));
    arm_fill_f32(0, B_+3, (fft_size-3));
    /*
    for (i = 3; i < fft_size; ++i)
    {
        A_[i] = 0;
        B_[i] = 0;
    }
    */
    create_filter(B_, A_, freq_filter);
}

void high_shelf(float K, float fc, float S, float fs, float complex* freq_filter)
{
    float A = pow(10, (K/20));
    float w0 = 2*pi*fc/fs;
    float alpha = sin(w0)/2 * sqrt((A+1/A)*(1/S-1)+2);
    
    float b0 = A*((A+1)+(A-1)*cos(w0)+2*sqrt(A)*alpha);
    float b1 = -2*A*((A-1)+(A+1)*cos(w0));
    float b2 = A*((A+1)+(A-1)*cos(w0)-2*sqrt(A)*alpha);
    float a0 = (A+1)-(A-1)*cos(w0)+2*sqrt(A)*alpha;
    float a1 = 2*((A-1)-(A+1)*cos(w0));
    float a2 = (A+1)-(A-1)*cos(w0)-2*sqrt(A)*alpha;
    
    float A_[fft_size];
    float B_[fft_size];
    int i = 0;
    A_[0] = a0;
    A_[1] = a1;
    A_[2] = a2;
    B_[0] = b0;
    B_[1] = b1;
    B_[2] = b2;
    arm_fill_f32(0, A_+3, (fft_size-3));
    arm_fill_f32(0, B_+3, (fft_size-3));
    /*
    for (i = 3; i < fft_size; ++i)
    {
        A_[i] = 0;
        B_[i] = 0;
    }
    */
    create_filter(B_, A_, freq_filter);
}

void construct_filterbank(float* gains, float* center_frequencies, float* band_widths, float fs,
		float complex filterbank[num_bands][n_freqs], float complex big_filter[n_freqs])
{
    int i = 0;
    float complex audio_buf[n_freqs];
    // arm_fill_f32(1, (float*)big_filter, n_freqs*2);
    for (i = 0; i < num_bands; ++i)
    {
        if (i == 0)
            low_shelf(gains[i], center_frequencies[i], band_widths[i], fs, filterbank[i]);
        else if (i == num_bands-1)
            high_shelf(gains[i], center_frequencies[i], band_widths[i], fs, filterbank[i]);
        else
            true_bell_filter(gains[i], center_frequencies[i], band_widths[i], fs, filterbank[i]);
        if (i % 2 == 0)
        {
        	if (i == 0)
                arm_copy_f32((float *)filterbank[i], (float *)(audio_buf), n_freqs*2);
        	else
        		arm_cmplx_mult_cmplx_f32((float*)big_filter,(float*)filterbank[i], (float*)audio_buf, n_freqs);
        }
        else
        	arm_cmplx_mult_cmplx_f32((float*)audio_buf,(float*)filterbank[i], (float*)big_filter, n_freqs);
    }

}

void construct_fast_filterbank(float* gains, float* center_frequencies, float* band_widths, float fs,
		arm_fir_instance_f32* filterbank[num_bands], float states[num_bands][fft_size*2-1])
{
    int i = 0;
    for (i = 0; i < num_bands; ++i)
    {
        if (i == 0)
            low_shelf_fast(gains[i], center_frequencies[i], band_widths[i], fs, filterbank[i], states[i]);
        else if (i == num_bands-1)
            high_shelf_fast(gains[i], center_frequencies[i], band_widths[i], fs, filterbank[i], states[i]);
        else
            true_bell_filter_fast(gains[i], center_frequencies[i], band_widths[i], fs, filterbank[i], states[i]);
    }
}

void hanning_window(float win_len, float* win_a)
{
    int n = 0;
    for (n = 0; n < win_len; ++n)
        win_a[n] = 0.5*(1-cos(2*pi*(n)/(win_len-1)));
}

void stft(float* audio_signal, float complex X[n_frames][n_freqs])
{
	arm_rfft_fast_instance_f32 varInstRfftF32;
	arm_rfft_fast_init_f32(&varInstRfftF32, fft_size);
    int n = 0;
    // int i = 0;
    float win_a[fft_size];
    hanning_window(fft_size, win_a);
    float buffer_in[fft_size];
    float complex buffer_out[fft_size];
    for(n = 0; n < n_frames; ++n)
    {
        int start_dx = n*hop;
        arm_mult_f32(audio_signal+start_dx, win_a, buffer_in, fft_size);
        // for (i = 0; i < fft_size; ++i)
        //    buffer_in[i] = audio_signal[start_dx+i]*win_a[i];
        arm_rfft_fast_f32(&varInstRfftF32, (float*)buffer_in, (float*)X[n], 0);
        ((float*)(X[n]))[fft_size] =  ((float*)(buffer_out))[1];
        ((float*)(X[n]))[fft_size+1] = 0;
        ((float*)(X[n]))[1] = 0;
        // fft(buffer_in, fft_size);
        // arm_copy_f32((float *)buffer_out, (float *)(X[n]), n_freqs*2);
        // for (i = 0; i < n_freqs; ++i)
        //     X[n][i] = buffer_out[i];
        	// X[n][i] = buffer_in[i];
    }
}

void istft(float complex X[n_frames][n_freqs], float* audio_signal)
{
	arm_rfft_fast_instance_f32 varInstRfftF32;
	arm_rfft_fast_init_f32(&varInstRfftF32, fft_size);
    int n = 0, i = 0;
    float win_a[fft_size];
    hanning_window(fft_size, win_a);
    float complex td_in[n_freqs];
    float td_out[fft_size];
    float norm[audio_len];
    float audio_buffer[fft_size];
    float norm_buffer[fft_size];
    float blanko[fft_size];
    arm_mult_f32(win_a, win_a, norm_buffer, fft_size);
    arm_fill_f32(0, norm, audio_len);
    // memset(norm, 0, audio_len*sizeof(float));
    for(n = 0; n < n_frames; ++n)
    {
    	// arm_copy_f32((float *)X[n], (float *)td_in, n_freqs*2);
        // for (i = 0; i < n_freqs; ++i)
        //     td_in[i] = X[n][i];
        // for (i = n_freqs; i < fft_size; ++i)
        //     td_in[i] = conj(td_in[(int)(fft_size-i)]);

        ((float*)(X[n]))[1] = ((float*)(X[n]))[fft_size];
        ((float*)(X[n]))[fft_size] = 0;
        arm_rfft_fast_f32(&varInstRfftF32, (float*)X[n], (float*)td_out, 1);
        // ift(td_sig, fft_size);
        int start_dx = n*hop;
        arm_mult_f32(td_out, win_a, audio_buffer, fft_size);
        arm_add_f32(audio_buffer, audio_signal+start_dx, blanko, fft_size);
        arm_copy_f32(blanko, audio_signal+start_dx, fft_size);
        arm_add_f32(norm_buffer, norm+start_dx, blanko, fft_size);
        arm_copy_f32(blanko, norm+start_dx, fft_size);
        // for (i = 0; i < fft_size; ++i)
        // {
            // audio_signal[start_dx+i] += (creal(td_sig[i])*win_a[i]);
        // 	audio_signal[start_dx+i] += (td_out[i])*win_a[i];
        //    norm[start_dx+i] += pow(win_a[i], 2);
        // }
    }

    for (i = 0; i < audio_len; ++i)
        if (norm[i] != 0)
        {
            audio_signal[i] /= norm[i];
        }
}

void filter_input(float* audio_signal, float* final_signal, float complex big_filter[n_freqs])
{
	uint32_t fb_start, fb_end;

    int n = 0, i = 0, j = 0;
    float complex X[n_frames][n_freqs];
	float complex tempo[n_freqs];
    fb_start = HAL_GetTick();
    stft(audio_signal, X);
    fb_end = HAL_GetTick();

    fb_start = HAL_GetTick();
    for (n = 0; n < n_frames; ++n)
    {
		arm_cmplx_mult_cmplx_f32((float*)X[n],(float*)big_filter, (float*)tempo, n_freqs);
		arm_copy_f32((float *)tempo, (float *)(X[n]), n_freqs*2);
		// for (i = 0; i < n_freqs; ++i)
		// {
		//    X[n][i] *= filterbank[j][i];
		// }
    }
    fb_end = HAL_GetTick();

    fb_start = HAL_GetTick();
    istft(X, final_signal);
    fb_end = HAL_GetTick();
}

void fast_filter(float* audio_signal, float* final_signal, float complex filterbank[num_bands][n_freqs])
{
	float eps = 1e-6;
	arm_rfft_fast_instance_f32 varInstRfftF32;
	arm_rfft_fast_init_f32(&varInstRfftF32, fft_size);
	float win_a[fft_size];
	float audio_buffer[fft_size];
	float complex X[n_freqs];
	float complex filt_buffer[n_freqs];
	int j = 0;
	// hanning_window(fft_size, win_a);

	// arm_mult_f32(audio_signal, win_a, audio_buffer, fft_size);
	// arm_rfft_fast_f32(&varInstRfftF32, (float*)audio_buffer, (float*)X, 0);
	arm_rfft_fast_f32(&varInstRfftF32, (float*)audio_signal, (float*)X, 0);

	for (j = 0; j < num_bands; ++j)
		{
			arm_cmplx_mult_cmplx_f32((float*)X,(float*)filterbank[j], (float*)filt_buffer, n_freqs);
			arm_copy_f32((float *)filt_buffer, (float *)(X), n_freqs*2);
		}
	// arm_rfft_fast_f32(&varInstRfftF32, (float*)X, (float*)audio_buffer, 1);
	arm_rfft_fast_f32(&varInstRfftF32, (float*)X, (float*)final_signal, 1);
	/*
	arm_mult_f32(audio_buffer, win_a, final_signal, fft_size);
	for (j = 0; j < fft_size; ++j)
	{
		final_signal[j] /= 2*pow	(win_a[j], 2)+eps;
	}
	*/

	/*
	for (j = 0; j < num_bands; ++j)
	{
		if (j % 2 == 0)
			arm_fir_f32(filterbank[j], audio_signal, final_signal, fft_size);
		else
			arm_fir_f32(filterbank[j], final_signal, audio_signal, fft_size);
	}
	*/
}

int initialize_with_chirp(float* sig, int start, int len, int c)
{
    int i = 0;
    for (i = 0; i < len; ++i)
    {
        sig[i] = sin(2.0*M_PI*(c/2.0*pow(((start+i)*1.0/FS), 2)))*VOLUME;
    }
    return start+i;
}

int save_to_file_continuous(float* final_signal, int aud_len) {
    static int counter = 0;
    int i = 0;
    FILE *fp;
    if (counter == 0)
    {
        fp = fopen("../tmp.txt","w");
        ++counter;
    }
    else
    {
        fp = fopen("../tmp.txt","a");
        fprintf(fp, " ");
    }
    for(i=0; i < aud_len; i++)
    {
        if (i < aud_len-1)
            fprintf(fp, "%f " ,final_signal[i]);
        else
            fprintf(fp, "%f" ,final_signal[i]);
    }
    fclose(fp);
    return 0;
}

/*
int save_to_file(float* final_signal, int audio_len, int b_dx) {
    FILE *fp;
    int i = 0;
    fp=fopen("../tmp.txt","w");
    for(i=b_dx; i < audio_len+b_dx; i++)
    {
        if (i < audio_len+b_dx-1)
            fprintf(fp, "%f " ,final_signal[i]);
        else
            fprintf(fp, "%f" ,final_signal[i]);
    }
    fclose(fp);
    return 0;
}
*/

/*
int save_to_file_filterbank(float complex** filterbank, int num_bands, float num_freqs) {
    FILE *fp_real;
    FILE *fp_complex;
    int i = 0, j=0;
    fp_complex=fopen("../tmp_bank_complex.txt", "w");
    fp_real=fopen("../tmp_bank_real.txt","w");
    for(i=0; i < num_bands; i++)
        for (j=0; j < num_freqs; ++j)
        {
            fprintf(fp_real, "%f " ,creal(filterbank[i][j]));
            fprintf(fp_complex, "%f " ,cimag(filterbank[i][j]));
        }
    fclose(fp_real);
    fclose(fp_complex);
    return 0;
}
*/
